(* The dictionary-passing translation of the Haskell code
   with typeclasses (typeclass_code.hs)
   into higher-order language (OCaml) without type classes.
*)

(* We always write signatures, even though they are inferred *)

(* The simplest type class: Show *)

(* Type-class declaration in Haskell corresponds to
   the dictionary data type declaration.
   Method's name becomes the label in the dictionary record.
*)
type 'a show = {show: 'a -> string}

(* The simplest instances *)
let show_bool : bool show = 
  {show = function
    | true  -> "True"
    | false -> "False"}

let show_int : int show =
  {show = string_of_int}

(* The first parametrically overloaded function *)
let print : 'a show -> 'a -> unit =
  fun {show=show} x -> print_endline (show x)

(* and its instantiation *)
let test_print : unit = print show_bool true


(* Another type class, whose methods follow the other two patterns
   (one method is binary)
   The Num type class has two methods; therefore, the corresponding
   dictionary record has two fields.
*)
type 'a num = 
    {fromInt: int -> 'a;    (* overloading on the result type *)
     add: 'a -> 'a -> 'a}

(* The two instances of Num *)
let num_bool : bool num = 
  {fromInt = (function 0 -> false | _ -> true);
   add = function 
     | true  -> fun _ -> true
     | false -> fun x -> x}

let num_int : int num = 
  {fromInt = (fun x -> x);
   add = Pervasives.(+)}

(* Another polymorphic function *)
let sum : 'a num -> 'a list -> 'a = 
  fun {fromInt = fromInt; add = add} ->
  fun ls -> List.fold_right add ls (fromInt 0)

let test_sum : int = sum num_int [1;3;5]
(* val test_sum : int = 9 *)

(* A polymorphic function that depends on two constraints *)
let print_incr : ('a show * 'a num) -> 'a -> unit =
  fun (show_dict, {fromInt=fromInt;add=(+)}) ->
  fun x -> print show_dict (x + fromInt 1)

(* Its instantiation. There are no longer constraints *)
let print_incr_int : int -> unit = fun x ->
  print_incr (show_int,num_int) x

let test_incr : unit = print_incr_int 4
(*
 5
 val test_incr : unit = ()
*)

(* An instance with a context *)
let show_list : 'a show -> 'a list show = 
  fun {show=show} ->
  {show = fun xs ->
   let rec go first = function
     | []   -> "]"
     | h::t ->  
         (* using show at a different type *)
         (if first then "" else ", ") ^ show h ^ go false t
   in "[" ^ go true xs
 }

let testls : string = 
  (show_list show_int).show [1;2;3]
(* val testls : string = "[1, 2, 3]" *)

(* Another useful class *)
type 'a eq = {eq: 'a -> 'a -> bool}

let eq_bool : bool eq = 
  {eq = fun x y -> match (x,y) with
  | (true,true) | (false,false) -> true
  | _ -> false}

let eq_int : int eq =
  {eq = Pervasives.(=)}

(* A type class with a super-class and default method *)
type 'a mul = {mul_super: 'a eq * 'a num;
               mul: 'a -> 'a -> 'a}

let mul_default : 'a eq * 'a num -> 'a mul =
  fun (({eq=eq},{fromInt=fromInt;add=(+)}) as super) ->
    (* Recursion is ordinary, not polymorphic *)
  {mul_super = super;
   mul = let rec loop x y = match () with
  | () when eq x (fromInt 0) -> fromInt 0
  | () when eq x (fromInt 1) -> y
  | () -> y + loop (x + (fromInt (-1))) y
  in loop
 }

let mul_bool : bool mul = mul_default (eq_bool,num_bool)

let mul_int : int mul =
  {mul_super=(eq_int,num_int); mul=Pervasives.( * )}

(* dot-product. There is only one constraint *)
let dot : 'a mul -> 'a list -> 'a list -> 'a =
  fun {mul_super=(eq,num);mul=mul} ->
  fun xs ys -> sum num @@ List.map2 mul xs ys

let test_dot : int = 
  dot mul_int [1;2;3] [4;5;6]
(* val test_dot : int = 32 *)

(* Auxiliary function for the example below *)
let rec replicate : int -> 'a -> 'a list = fun n x ->
  if n <= 0 then [] else x::replicate (n-1) x

(* Polymorphic recursion. The signature is mandatory here *)
let rec print_nested : 'a. 'a show -> int -> 'a -> unit =
  fun show_dict -> function 
    | 0 -> fun x -> print show_dict x
    | n -> fun x -> 
        print_nested (show_list show_dict) (n-1) (replicate n x)

let test_nested =
  let n = read_int () in
  print_nested show_int n 5

(* If entered 4

[[[[5, 5, 5, 5], [5, 5, 5, 5], [5, 5, 5, 5]], [[5, 5, 5, 5], [5, 5, 5, 5], [5, 5, 5, 5]]]]
val test_nested : unit = ()
*)

